/* $Id: redd.c,v 0.9 1999/12/28 11:58:07 kjc Exp kjc $ */
/*
 * Copyright (C) 1997, 1998
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <net/if.h>

#include <altq/altq.h>
#include <altq/altq_red.h>

#define RED_DEV	"/dev/altq/red"

int fd = -1;
int debug = 0;
int use_ecn = 0;
int use_flowvalve = 0;
int config = 0;
int maxqlen = 60;
int bandwidth = 10000000;	/* 10Mbps */
int packet_size = 1000;		/* average packet size: 1000 bytes */
int weight = 0;			/* 0: use default */
int inv_pmax = 0;		/* 0: use default */
int th_min = 0;			/* 0: use default */
int th_max = 0;			/* 0: use default */

static void usage(void)
{
	fprintf(stderr, "usage: redd [-dex] [-l maxqlen]\n"
		"    [-w weight] [-p inv_pmax] [-t th_min] [-T th_max]\n"
		"    [-b bandwidth] [-s packet_size] if_name\n");
	fprintf(stderr, "	-d: debug\n");
	fprintf(stderr, "	-e: use ECN (packet marking)\n");
	fprintf(stderr, "	-x: use flowvalve\n");
	fprintf(stderr, "	-l maxqlen: set max queue length\n");
	fprintf(stderr, "	-w weight: set red weight\n");
	fprintf(stderr, "	-p inv_pmax: set red drop probability\n");
	fprintf(stderr, "	-t th_min: set red min threshold\n");
	fprintf(stderr, "	-T th_max: set red max threshold\n");
	fprintf(stderr, "	-b bandwidth\n");
	fprintf(stderr, "	-s packet_size\n");
	exit(1);
}

static void
sigint_handler(int sig)
{
	/* do nothing */
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	int ch;
	struct red_interface iface;
	char *if_name, *cp;

	while ((ch = getopt(argc, argv, "dexl:w:p:t:T:b:s:")) != EOF) {
		switch (ch) {
		case 'd':
			debug = 1;
			break;
		case 'e':
			use_ecn = 1;
			config = 1;
			break;
		case 'x':
			use_flowvalve = 1;
			config = 1;
			break;
		case 'l':
			maxqlen = atoi(optarg);
			config = 1;
			break;
		case 'w':
			weight = atoi(optarg);
			config = 1;
			break;
		case 'p':
			inv_pmax = atoi(optarg);
			config = 1;
			break;
		case 't':
			th_min = atoi(optarg);
			config = 1;
			break;
		case 'T':
			th_max = atoi(optarg);
			config = 1;
			break;
		case 'b':
			cp = NULL;
			bandwidth = strtol(optarg, &cp, 0);
			if (cp != NULL) {
				if (*cp == 'K' || *cp == 'k')
					bandwidth *= 1000;
				else if (*cp == 'M' || *cp == 'm')
					bandwidth *= 1000000;
			}
			config = 1;
			break;
		case 's':
			cp = NULL;
			packet_size = strtol(optarg, &cp, 0);
			if (cp != NULL && (*cp == 'K' || *cp == 'k'))
				packet_size *= 1024;
			config = 1;
			break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc > 0) {
		if_name = argv[0];
	}
	else {
		if (argc == 0)
			fprintf(stderr, "missing interface name\n");
		usage();
	}

	signal(SIGINT, sigint_handler);
	signal(SIGTERM, sigint_handler);

	openlog("redd", (debug ? (LOG_PID|LOG_PERROR) : LOG_PID), LOG_DAEMON);

	if (!debug)
		daemon(0, 0);

	if((fd = open(RED_DEV, O_RDWR)) < 0) {
		syslog(LOG_ERR, "can't open %s: %m", RED_DEV);
		exit(1);
	}

	if (debug) {
		printf("redd on interface %s\n", if_name);
		printf("  set queue limit to %d\n", maxqlen);
		printf("  set bandwidth to %d bps\n", bandwidth);
		printf("  set packet_size to %d bytes\n", packet_size);
		printf("  set thresholds [%d:%d], inv_pmax to %d\n",
		       th_min, th_max, inv_pmax);
		if (use_flowvalve)
		    printf("  flowvalve enabled\n");
	}

	iface.red_ifname[IFNAMSIZ-1] = '\0';
	strncpy(iface.red_ifname, if_name, IFNAMSIZ-1);
    
	if (ioctl(fd, RED_IF_ATTACH, &iface) < 0) {
		syslog(LOG_ERR, "RED_IF_ATTACH for %s: %m", if_name);
		exit(1);
	}

	if (config) {
		struct red_conf conf;

		if (weight != 0) {
			/* check if weight is power of 2 */
			int i, w;
		
			w = weight;
			for (i = 0; w > 1; i++)
				w = w >> 1;
			w = 1 << i;
			if (weight != w) {
				syslog(LOG_ERR, "wrong weight %d for %s: %m",
				       weight, if_name);
				exit(1);
			}
		}

		bzero(&conf, sizeof(struct red_conf));
		conf.iface = iface;
		conf.red_weight = weight;
		conf.red_inv_pmax = inv_pmax;
		conf.red_thmin = th_min;
		conf.red_thmax = th_max;
		conf.red_limit = maxqlen;
		conf.red_pkttime = packet_size * 8 * 1000 / (bandwidth / 1000);
		conf.red_flags = use_ecn ? REDF_ECN : 0;
		if (use_flowvalve)
			conf.red_flags |= REDF_FLOWVALVE;
	
		if (debug)
			printf("  packet time set to %d usec\n",
			       conf.red_pkttime);
	    
		if (ioctl(fd, RED_CONFIG, &conf) < 0) {
			syslog(LOG_ERR, "RED_CONFIG: %m");
			exit(1);
		}
	}

	if (ioctl(fd, RED_ENABLE, &iface) < 0) {
	    syslog(LOG_ERR, "RED_ENABLE: %m");
	    exit(1);
	}

	/* wait for signals and then terminate */
	pause();

	if (debug)
		printf("Got signal.  Exiting...\n");

	if (ioctl(fd, RED_DISABLE, &iface) < 0)
	    syslog(LOG_ERR, "RED_DISABLE: %m");

	if (ioctl(fd, RED_IF_DETACH, &iface) < 0)
		syslog(LOG_ERR, "RED_IF_DETACH: %m");

	close(fd);
	closelog();

	return 0;
}
