/* $Id: riod.c,v 1.4 1999/12/28 11:58:24 kjc Exp kjc $ */
/*
 * Copyright (C) 1997, 1998
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <syslog.h>

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <net/if.h>

#include <altq/altq.h>
#include <altq/altq_red.h>
#include <altq/altq_rio.h>

#define RIO_DEV	"/dev/altq/rio"

int fd = -1;
int debug = 0;
int use_ecn = 0;
int config = 0;
int maxqlen = 60;
int bandwidth = 10000000;	/* 10Mbps */
int packet_size = 1000;		/* average packet size: 1000 bytes */
int weight = 0;			/* 0: use default */
int lo_inv_pmax = 0;		/* 0: use default */
int lo_th_min = 0;		/* 0: use default */
int lo_th_max = 0;		/* 0: use default */
int med_inv_pmax = 0;		/* 0: use default */
int med_th_min = 0;		/* 0: use default */
int med_th_max = 0;		/* 0: use default */
int hi_inv_pmax = 0;		/* 0: use default */
int hi_th_min = 0;		/* 0: use default */
int hi_th_max = 0;		/* 0: use default */
int clear_field = 0;

static void usage(void)
{
	fprintf(stderr, "usage: riod [-ademox] [-l maxqlen] [-w weight]\n"
		"    [-p lo_inv_pmax] [-t lo_th_min] [-T lo_th_max]\n"
		"    [-q med_inv_pmax] [-m med_th_min] [-M med_th_max]\n"
		"    [-r hi_inv_pmax] [-h hi_th_min] [-H hi_th_max]\n"
		"    [-b bandwidth] [-s packet_size] if_name\n");
	fprintf(stderr, "		-d: debug\n");
	fprintf(stderr, "		-e: use ECN (packet marking)\n");
	fprintf(stderr, "		-x: clear ds-field\n");
	fprintf(stderr, "		-l maxqlen: set max queue length\n");
	fprintf(stderr, "		-w weight: set red weight\n");
	fprintf(stderr, "		-p inv_pmax: drop probability\n");
	fprintf(stderr, "		-t th_min: min threshold\n");
	fprintf(stderr, "		-T th_max: max threshold\n");
	fprintf(stderr, "		-q inv_pmax: for medium dp\n");
	fprintf(stderr, "		-m th_min: for medium dp\n");
	fprintf(stderr, "		-M th_max: for medium dp\n");
	fprintf(stderr, "		-r inv_pmax: for high dp\n");
	fprintf(stderr, "		-h th_min: for high dp\n");
	fprintf(stderr, "		-H th_max: for high dp\n");
	fprintf(stderr, "		-b bandwidth\n");
	fprintf(stderr, "		-s packet_size\n");
	exit(1);
}

static void
sigint_handler(int sig)
{
	/* do nothing */
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	int ch;
	struct rio_interface iface;
	char *if_name, *cp;

	while ((ch = getopt(argc, argv, "dexl:w:p:t:T:q:m:M:r:h:H:b:s:")) != EOF) {
		switch (ch) {
		case 'd':
			debug = 1;
			break;
		case 'e':
			use_ecn = 1;
			config = 1;
			break;
		case 'x':
			clear_field = 1;
			config = 1;
			break;
		case 'l':
			maxqlen = atoi(optarg);
			config = 1;
			break;
		case 'w':
			weight = atoi(optarg);
			config = 1;
			break;
		case 'p':
			lo_inv_pmax = atoi(optarg);
			config = 1;
			break;
		case 't':
			lo_th_min = atoi(optarg);
			config = 1;
			break;
		case 'T':
			lo_th_max = atoi(optarg);
			config = 1;
			break;
		case 'q':
			med_inv_pmax = atoi(optarg);
			config = 1;
			break;
		case 'm':
			med_th_min = atoi(optarg);
			config = 1;
			break;
		case 'M':
			med_th_max = atoi(optarg);
			config = 1;
			break;
		case 'r':
			hi_inv_pmax = atoi(optarg);
			config = 1;
			break;
		case 'h':
			hi_th_min = atoi(optarg);
			config = 1;
			break;
		case 'H':
			hi_th_max = atoi(optarg);
			config = 1;
			break;
		case 'b':
			cp = NULL;
			bandwidth = strtol(optarg, &cp, 0);
			if (cp != NULL) {
				if (*cp == 'K' || *cp == 'k')
					bandwidth *= 1000;
				else if (*cp == 'M' || *cp == 'm')
					bandwidth *= 1000000;
			}
			config = 1;
			break;
		case 's':
			cp = NULL;
			packet_size = strtol(optarg, &cp, 0);
			if (cp != NULL && (*cp == 'K' || *cp == 'k'))
				packet_size *= 1024;
			config = 1;
			break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc > 0) {
		if_name = argv[0];
	}
	else {
		if (argc == 0)
			fprintf(stderr, "missing interface name\n");
		usage();
	}

	signal(SIGINT, sigint_handler);
	signal(SIGTERM, sigint_handler);

	openlog("redd", (debug ? (LOG_PID|LOG_PERROR) : LOG_PID), LOG_DAEMON);

	if (!debug)
		daemon(0, 0);

	if((fd = open(RIO_DEV, O_RDWR)) < 0) {
		syslog(LOG_ERR, "can't open %s: %m", RIO_DEV);
		exit(1);
	}

	if (debug) {
		printf("riod on interface %s\n", if_name);
		printf("  set queue limit to %d\n", maxqlen);
		printf("  set bandwidth to %d bps\n", bandwidth);
		printf("  set packet_size to %d bytes\n", packet_size);
	}

	iface.rio_ifname[IFNAMSIZ-1] = '\0';
	strncpy(iface.rio_ifname, if_name, IFNAMSIZ-1);
    
	if (ioctl(fd, RIO_IF_ATTACH, &iface) < 0) {
		syslog(LOG_ERR, "RIO_IF_ATTACH for %s: %m", if_name);
		exit(1);
	}

	if (config) {
		struct rio_conf conf;

		if (weight != 0) {
			/* check if weight is power of 2 */
			int i, w;
		
			w = weight;
			for (i = 0; w > 1; i++)
				w = w >> 1;
			w = 1 << i;
			if (weight != w) {
				syslog(LOG_ERR, "wrong weight %d for %s: %m",
				       weight, if_name);
				exit(1);
			}
		}

		bzero(&conf, sizeof(struct rio_conf));
		conf.iface = iface;
		conf.q_params[0].inv_pmax = lo_inv_pmax;
		conf.q_params[0].th_min = lo_th_min;
		conf.q_params[0].th_max = lo_th_max;
		conf.q_params[1].inv_pmax = med_inv_pmax;
		conf.q_params[1].th_min = med_th_min;
		conf.q_params[1].th_max = med_th_max;
		conf.q_params[2].inv_pmax = hi_inv_pmax;
		conf.q_params[2].th_min = hi_th_min;
		conf.q_params[2].th_max = hi_th_max;
		conf.rio_weight = weight;
		conf.rio_limit = maxqlen;
		conf.rio_pkttime = packet_size * 8 * 1000 / (bandwidth / 1000);
		conf.rio_flags = use_ecn ? RIOF_ECN : 0;
		if (clear_field)
			conf.rio_flags |= RIOF_CLEARDSCP;
		
		if (ioctl(fd, RIO_CONFIG, &conf) < 0) {
			syslog(LOG_ERR, "RIO_CONFIG: %m");
			exit(1);
		}

		if (debug)
			printf("  packet time %d usec\n", conf.rio_pkttime);
	}

	if (ioctl(fd, RIO_ENABLE, &iface) < 0) {
	    syslog(LOG_ERR, "RIO_ENABLE: %m");
	    exit(1);
	}

	/* wait for signals and then terminate */
	pause();

	if (debug)
		printf("Got signal.  Exiting...\n");

	if (ioctl(fd, RIO_DISABLE, &iface) < 0)
	    syslog(LOG_ERR, "RIO_DISABLE: %m");

	if (ioctl(fd, RIO_IF_DETACH, &iface) < 0)
		syslog(LOG_ERR, "RIO_IF_DETACH: %m");

	close(fd);
	closelog();

	return 0;
}
