/*
 * Copyright (C) 1997, 1999
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: wfqd.c,v 0.5 1999/12/28 11:58:37 kjc Exp kjc $
 */
/*
 *  March 27, 1997.  Written by Hiroshi Kyusojin of Keio University
 *  (kyu@mt.cs.keio.ac.jp). 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <err.h>
#include <ctype.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <netinet/in.h>
#include <altq/altq.h>
#include <altq/altq_wfq.h>
#include <arpa/inet.h>

#define WFQ_DEV		"/dev/altq/wfq"
#ifndef IFNAMSIZ
#define IFNAMSIZ	16
#endif
#define PROMPT		"> "
#define QUIT		"quit"
#define ENABLE		"enable"
#define DISABLE		"disable"
#define GETSTATS	"getstats"
#define SETWEIGHT	"setweight"
#define GETQID		"getqid"
#define IFLIST_SIZE	16

int debug = 0;
int interactive = 0;
int hash_policy = WFQ_HASH_DSTADDR;
int nqueues = 0;	/* use default */
int qlimit = 0;		/* use default */

/*
 * prototype
 */
int get_ifname(char *, char *);
int do_command(int, char *);

static void usage(void)
{
	fprintf(stderr, "usage: wfqd [-DFSdi] [-n nqueues] [-l limit] [if_name ...]\n");
	fprintf(stderr, "	-D: dst addr hash policy (default)\n");
	fprintf(stderr, "	-F: full hash policy\n");
	fprintf(stderr, "	-S: src port hash policy\n");
	fprintf(stderr, "	-d: debug (do not daemonize)\n");
	fprintf(stderr, "	-i: interactive\n");
	fprintf(stderr, "	-l: queue size limit in bytes\n");
	fprintf(stderr, "	-n: the number of queues\n");
	exit(1);
}

static void
sigint_handler(int sig)
{
	/* do nothing */
}

int main(int argc, char *argv[])
{
	int fd, ch, i, if_idx;
	char *cp, *if_names[IFLIST_SIZE], buf[256];

	while ((ch = getopt(argc, argv, "DFSdil:n:")) != EOF) {
		switch (ch) {
		case 'D':
			hash_policy = WFQ_HASH_DSTADDR;
			break;
		case 'F':
			hash_policy = WFQ_HASH_FULL;
			break;
		case 'S':
			hash_policy = WFQ_HASH_SRCPORT;
			break;
		case 'd':
			debug = 1;
			break;
		case 'i':
			interactive = 1;
			break;
		case 'l':
			cp = NULL;
			qlimit = strtol(optarg, &cp, 0);
			if (cp != NULL) {
				if (*cp == 'K' || *cp == 'k')
					qlimit *= 1024;
				else if (*cp == 'M' || *cp == 'm')
					qlimit *= 1024 * 1024;
			}
			if (qlimit < 1500) {
				fprintf(stderr, "qlimit:%d Bytes too small!\n",
					qlimit);
				usage();
			}
			break;
		case 'n':
			nqueues = atoi(optarg);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (argc > 0) {

		for (if_idx = 0; if_idx < IFLIST_SIZE && argc > 0; if_idx++) {
			if_names[if_idx] = argv[0];
			argc--;
			argv++;
		}
		if (if_idx == IFLIST_SIZE)
			errx(1, "wfqd: too many interfaces specified!");
	}
	else if (!interactive) {
		fprintf(stderr, "missing interface name!\n");
		usage();
	}

	signal(SIGINT, sigint_handler);
	signal(SIGTERM, sigint_handler);

	/*
	 * open file and get a file descriptor
	 */
	if ((fd = open(WFQ_DEV, O_RDONLY)) < 0)
		err(1, "Can't open wfq device file");

	for (i = 0; i < if_idx; i++) {
		sprintf(buf, "%s %s %d", if_names[i], ENABLE, nqueues);
		do_command(fd, buf);
	}

	if (interactive) {
		
		while(1){
			printf("%s", PROMPT);
			fgets(buf, sizeof(buf), stdin);
    
			/*
			 * execute command
			 */
			if (do_command(fd, buf) < 0)
				break;
		}
	}
	else {
		if (!debug)
			daemon(0, 0);

		/* wait for signals and then terminate */
		pause();

		if (debug)
			printf("Got signal.  Exiting...\n");
	}

	for (i = 0; i < if_idx; i++) {
		sprintf(buf, "%s %s", if_names[i], DISABLE);
		do_command(fd, buf);
	}

	close(fd);
	return (0);
}


int get_ifname(char *ifname, char *p)
{
	int i;

	for (i = 0; i < IFNAMSIZ - 1 && *p != ' ' && *p != '\t' && *p != '\n' ; i++)
		ifname[i] = *p++;
	ifname[i] = '\0';

	return i;
}

int do_command(int fd, char *buf)
{
	struct wfq_interface iface;
	int len;
	char *p, ifname[IFNAMSIZ];

	for (p = buf; *p == ' ' || *p == '\t'; p++)
		;
	if (*p == '\n')
		return 0;

	if (!strncmp(QUIT, p, strlen(QUIT)))
		return -1;

	/*
	 * get interface name
	 */
	if ((len = get_ifname(ifname, p)) == IFNAMSIZ - 1){
		fprintf(stderr, "too long interface name : %s\n", ifname);
		return 0;
	}
	strcpy(iface.wfq_ifacename, ifname);

	p += len;
	while(*p == ' ' || *p == '\t')
		p++;

	if (!strncmp(ENABLE, p, strlen(ENABLE))){
		int n;

		p += strlen(ENABLE);
		while (*p == ' ' || *p == '\t')
			p++;
		for (n = 0; '0' <= *p && *p <= '9'; p++)
			n = n * 10 + *p - '0';
		if (n != 0)
			nqueues = n;

		if (ioctl(fd, WFQ_IF_ATTACH, &iface) < 0)
			err(1, "ioctl WFQ_IF_ATTACH");

		if (hash_policy != 0 || nqueues != 0 || qlimit != 0) {
			struct wfq_conf conf;

			conf.iface = iface;
			conf.nqueues = nqueues;
			conf.qlimit = qlimit;
			conf.hash_policy = hash_policy;
			if (ioctl(fd, WFQ_CONFIG, &conf) < 0)
				err(1, "ioctl WFQ_CONFIG");
		}
		
		if (ioctl(fd, WFQ_ENABLE, &iface) < 0)
			err(1, "ioctl WFQ_ENABLE");
	}
  	else if (!strncmp(DISABLE, p, strlen(DISABLE))){
		if (ioctl(fd, WFQ_DISABLE, &iface) < 0)
			err(1, "ioctl WFQ_DISABLE");
		if (ioctl(fd, WFQ_IF_DETACH, &iface) < 0)
			err(1, "ioctl WFQ_IF_DETACH");
	}
	else if (!strncmp(GETQID, p, strlen(GETQID))){
		struct wfq_getqid gqid;
		struct flowinfo_in *fi;
		u_long tmp, val = 0;

		bzero(&gqid, sizeof(gqid));
		gqid.iface = iface;

		p += strlen(GETQID);
		while (*p == ' ' || *p == '\t')
			p++;

		do{
			for (tmp = 0;  '0' <= *p && *p <= '9'; p++)
				tmp = tmp * 10 + *p - '0';
			val = (val << 8) | tmp;
		} while (*p++ == '.');
		fi = (struct flowinfo_in *)&gqid.flow;
		fi->fi_family = AF_INET;
		fi->fi_dst.s_addr = htonl(val);
		if(ioctl(fd, WFQ_GET_QID, &gqid) < 0)
			err(1, "ioctl WFQ_GET_QID");

		printf("queue ID = %ld for dest (%s)\n",
		       gqid.qid, inet_ntoa(fi->fi_dst));
	}
	else if (!strncmp(SETWEIGHT, p, strlen(SETWEIGHT))) {
		struct wfq_setweight sw;
    
		sw.iface = iface;
    
		p += strlen(SETWEIGHT);
		while (*p == ' ' || *p == '\t')
			p++;
    
		for (sw.qid = 0; '0' <= *p && *p <= '9'; p++)
			sw.qid = sw.qid * 10 + *p - '0';
    
		while(*p == ' ' || *p == '\t')
			p++;
    
		for (sw.weight = 0; '0' <= *p && *p <= '9'; p++)
			sw.weight = sw.weight * 10 + *p - '0';
		if (ioctl(fd, WFQ_SET_WEIGHT, &sw) < 0)
			err(1, "ioctl WFQ_SET_WEIGHT");
	}
	else if (!strncmp(GETSTATS, p, strlen(GETSTATS))){
		struct wfq_getstats gs;
    
		gs.iface = iface;
    
		p += strlen(GETSTATS);
		while (*p == ' ' || *p == '\t')
			p++;
    
		for (gs.qid = 0; '0' <= *p && *p <= '9'; p++)
			gs.qid = gs.qid * 10 + *p - '0';

		if (ioctl(fd, WFQ_GET_STATS, &gs) < 0)
			err(1, "ioctl WFQ_GET_STATS");
    
		printf("Bytes  = %d\n", gs.stats.bytes);
		printf("Weight = %d\n", gs.stats.weight);
		printf("Sent : packets = %llu, bytes = %llu\n",
		       gs.stats.xmit_cnt.packets, gs.stats.xmit_cnt.bytes);
		printf("Drop : packets = %llu, bytes = %llu\n",
		       gs.stats.drop_cnt.packets, gs.stats.drop_cnt.bytes);
	}

	else{
		fprintf(stderr, "Usage : interface %s [nqueues]\n", ENABLE);
		fprintf(stderr, "        interface %s\n", DISABLE);
		fprintf(stderr, "        interface %s ipaddr(XX.XX.XX.XX)\n", GETQID);
		fprintf(stderr, "        interface %s qid\n", GETSTATS);
		fprintf(stderr, "        interface %s qid weight\n", SETWEIGHT);
	}

	return 0;
}
