/* $Id: wfqstat.c,v 1.2 1999/12/28 11:58:37 kjc Exp kjc $ */
/*
 * Copyright (C) 1999
 *	Sony Computer Science Laboratory Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <err.h>
#ifndef NO_CURSES
#include <curses.h>
#endif

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <net/if.h>

#include <altq/altq.h>
#include <altq/altq_wfq.h>

#define	WFQ_DEVICE	"/dev/altq/wfq"
#define NTOP		10

struct wfqinfo {
	int qid;
	queue_stats stats;
	u_quad_t last_bytes;
	double mbps;
};

int fd = -1;
char *if_name = NULL;
int interval = 5;
int ntop = NTOP;

static void usage(void)
{
	fprintf(stderr, "usage: wfqstat [-i interval] [-n ntop] interface\n");
	exit(1);
}

static void
sigint_handler(int sig)
{
#ifndef NO_CURSES
	endwin();
#endif
	fprintf(stderr, "Exiting on signal %d\n", sig);

	close(fd);
	exit(0);
}

int main(int argc, char **argv)
{
	struct wfq_getstats wfq_stats;
	struct timeval cur_time, last_time;
	u_quad_t xmit_bytes, last_bytes;
	int msec, i, j, k, nqueues;
	char *if_name;
	int ch;
	struct wfqinfo *qinfo, **top;

	while ((ch = getopt(argc, argv, "i:n:")) != EOF) {
		switch (ch) {
		case 'i':
			interval = atoi(optarg);
			break;
		case 'n':
			ntop = atoi(optarg);
			break;
		}
	}
	argc -= optind;
	argv += optind;
	if (argc > 0) {
		if_name = argv[0];
	}
	else {
		if (argc == 0)
			fprintf(stderr, "missing interface name\n");
		usage();
	}

	signal(SIGINT, sigint_handler);
	signal(SIGTERM, sigint_handler);

	if ((fd = open(WFQ_DEVICE, O_RDONLY)) < 0)
		err(1, "wfq open");

	if (strlen(if_name) > IFNAMSIZ)
		err(1, "interface name %s too long!\n", if_name);

	wfq_stats.iface.wfq_ifacename[IFNAMSIZ-1] = '\0';
	strncpy(wfq_stats.iface.wfq_ifacename, if_name, IFNAMSIZ-1);

	/*
	 * first, find out how many queues are available
	 */
	for (i = 0; i < MAX_QSIZE; i++) {
		wfq_stats.qid = i;
		if (ioctl(fd, WFQ_GET_STATS, &wfq_stats) < 0)
			break;
	}
	nqueues = i;
	printf("wfq on %s: %d queues are used\n", if_name, nqueues);

	if ((qinfo = malloc(nqueues * sizeof(struct wfqinfo))) == NULL)
		err(1, "malloc failed!");
	if ((top = malloc(ntop * sizeof(struct wfqinfo *))) == NULL)
		err(1, "malloc failed!");

#ifndef NO_CURSES
	sleep(2);  /* wait a bit before clearing the screen */

	initscr();
#endif

	gettimeofday(&last_time, NULL);
	last_time.tv_sec -= interval;
	last_bytes = 0;

	while (1) {
		for (j = 0; j < ntop; j++)
			top[j] = NULL;

		for (i = 0; i < nqueues; i++) {
			wfq_stats.qid = i;
			if (ioctl(fd, WFQ_GET_STATS, &wfq_stats) < 0)
				err(1, "ioctl WFQ_GET_STATS");

			qinfo[i].qid = i;
			qinfo[i].stats = wfq_stats.stats;
		}

		gettimeofday(&cur_time, NULL);
		msec = (cur_time.tv_sec - last_time.tv_sec)*1000 +
			(cur_time.tv_usec - last_time.tv_usec)/1000;
		last_time = cur_time;

		/*
		 * calculate the throughput of each queue
		 */
		for (i = 0; i < nqueues; i++) {
			xmit_bytes = qinfo[i].stats.xmit_cnt.bytes - 
				qinfo[i].last_bytes;
			qinfo[i].mbps = (double)xmit_bytes * 8.0
				/ (double)msec * 1000.0 / 1000.0 / 1000.0;
			qinfo[i].last_bytes = qinfo[i].stats.xmit_cnt.bytes;

			for (j = 0; j < ntop; j++) {
				if (top[j] == NULL) {
					top[j] = &qinfo[i];
					break;
				}
				if (top[j]->mbps < qinfo[i].mbps ||
				    (top[j]->mbps == qinfo[i].mbps &&
				     top[j]->stats.xmit_cnt.packets <
				     qinfo[i].stats.xmit_cnt.packets)) {
					for (k = ntop-1; k > j; k--)
						top[k] = top[k-1];
					top[j] = &qinfo[i];
					break;
				}
			}
		}

		/*
		 * display top
		 */
		printf("[QID] WEIGHT QSIZE(KB) SENT(pkts)     (KB)       DROP(pkts)     (KB)     Mbps\n");

		for (j = 0; j < ntop; j++) {
			if (top[j] != NULL)
				printf("[%4d] %4d %4d %10llu %14llu %10llu %14llu %8.2f\n",
				       top[j]->qid,
				       top[j]->stats.weight,
				       top[j]->stats.bytes / 1024,
				       top[j]->stats.xmit_cnt.packets, 
				       top[j]->stats.xmit_cnt.bytes / 1024, 
				       top[j]->stats.drop_cnt.packets, 
				       top[j]->stats.drop_cnt.bytes / 1024,
				       top[j]->mbps);
			else
				printf("\n");
		}	
#ifndef NO_CURSES
		refresh();
		mvcur(ntop+1, 0, 0, 0);
#endif

		sleep(interval);
	}

	return (0);
}
